from unittest.mock import patch

from django_kafka.models import KafkaSkipMixin
from django_kafka.tests.models import AbstractModelTestCase


class KafkaSkipModelTestCase(AbstractModelTestCase):
    abstract_model = KafkaSkipMixin

    @patch("django_kafka.models.super")
    def test_save_update_fields_not_in_kwargs(self, super_mock):
        """
        kafka_skip should be set to False when update_fields is not provided
        """

        instance = self.model(pk=1, kafka_skip=True)

        instance.save()

        # save sets kafka_skip value to False
        self.assertFalse(instance.kafka_skip)
        # didn't forget to call super
        super_mock.assert_called_once_with()
        # no update_fields set
        super_mock.return_value.save.assert_called_once_with(
            force_insert=False,
            force_update=False,
            using=None,
            update_fields=None,
        )

    @patch("django_kafka.models.super")
    def test_save_update_fields_in_kwargs_without_kafka_skip(self, super_mock):
        """
        kafka_skip should be set to False when update_fields does not contain kafka_skip
        """
        instance = self.model(pk=1, kafka_skip=True)

        update_fields = ["some_fields"]

        instance.save(update_fields=update_fields)

        # save sets kafka_skip value to False
        self.assertFalse(instance.kafka_skip)
        # didn't forget to call super
        super_mock.assert_called_once_with()
        # kafka_skip added to update_fields
        super_mock.return_value.save.assert_called_once_with(
            force_insert=False,
            force_update=False,
            using=None,
            update_fields=["kafka_skip", *update_fields],
        )

    @patch("django_kafka.models.super")
    def test_save_update_fields_in_kwargs_with_kafka_skip(self, super_mock):
        """
        kafka_skip should not be changed if provided in update_fields
        """
        update_fields = ["some_field", "kafka_skip"]

        instance = self.model(pk=1, kafka_skip=True)

        instance.save(update_fields=update_fields)

        # save does not change kafka_skip value
        self.assertTrue(instance.kafka_skip)
        # didn't forget to call super
        super_mock.assert_called_once_with()
        # update_fields are not changed
        super_mock.return_value.save.assert_called_once_with(
            force_insert=False,
            force_update=False,
            using=None,
            update_fields=update_fields,
        )

    @patch("django_kafka.models.super")
    def test_queryset_update_sets_kafka_skip(self, super_mock):
        """
        `update` method should automatically set `kafka_skip=False`
          if `kafka_skip` is not provided in kwargs.
        """
        # update_kwargs does not contain kafka_skip
        update_kwargs = {"some_field": "some_value"}

        self.model.objects.update(**update_kwargs)

        # kafka_skip=False was added to the update_kwargs
        super_mock.return_value.update.assert_called_once_with(
            **{"kafka_skip": False, **update_kwargs},
        )

    @patch("django_kafka.models.super")
    def test_queryset_update_does_not_override_kafka_skip(self, super_mock):
        """
        `update` method should not change `kafka_skip`
          if `kafka_skip` is provided in kwargs
        """
        # update_kwargs contains kafka_skip
        update_kwargs = {"kafka_skip": True, "some_field": "some_value"}

        self.model.objects.update(**update_kwargs)

        # kafka_skip is not changed, update_kwargs are not changed
        super_mock.return_value.update.assert_called_once_with(**update_kwargs)



from unittest.mock import patch

from django.test import TestCase, override_settings

from django_kafka.conf import DEFAULTS, SETTINGS_KEY, settings


class SettingsTestCase(TestCase):
    settings_keys = (
        "CLIENT_ID",
        "ERROR_HANDLER",
        "GLOBAL_CONFIG",
        "PRODUCER_CONFIG",
        "CONSUMER_CONFIG",
        "POLLING_FREQUENCY",
        "SCHEMA_REGISTRY",
    )

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_defaults(self, mock_consumer_client):
        # make sure defaults are assigned
        for key in self.settings_keys:
            self.assertEqual(getattr(settings, key), DEFAULTS[key])

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_user_settings(self, mock_consumer_client):
        # make sure settings defined by user pulled up
        user_settings = {
            "CLIENT_ID": "client-id",
            "ERROR_HANDLER": "error.handler.class",
            "GLOBAL_CONFIG": {"bootstrap.servers": "kafka1"},
            "PRODUCER_CONFIG": {
                "enable.idempotence": True,
            },
            "CONSUMER_CONFIG": {
                "group.id": "group-1",
            },
            "POLLING_FREQUENCY": 0.5,
            "SCHEMA_REGISTRY": {
                "url": "https://schema-registry",
            },
        }
        with override_settings(**{SETTINGS_KEY: user_settings}):
            for key in self.settings_keys:
                self.assertEqual(getattr(settings, key), user_settings[key])



from django.test import TestCase

from django_kafka.consumer import Consumer
from django_kafka.registry import ConsumersRegistry


class ConsumersRegistryTestCase(TestCase):
    def test_decorator_adds_to_registry(self):
        class ConsumerA(Consumer):
            pass

        class ConsumerB(Consumer):
            pass

        registry = ConsumersRegistry()

        self.assertIs(registry()(ConsumerA), ConsumerA)
        self.assertIs(registry()(ConsumerB), ConsumerB)

        key_a = registry.get_key(ConsumerA)
        self.assertIs(registry[key_a], ConsumerA)

        key_b = registry.get_key(ConsumerB)
        self.assertIs(registry[key_b], ConsumerB)

    def test_iter_returns_keys(self):
        class ConsumerA(Consumer):
            pass

        class ConsumerB(Consumer):
            pass

        registry = ConsumersRegistry()

        registry()(ConsumerA)
        registry()(ConsumerB)

        self.assertListEqual(
            list(registry),
            [registry.get_key(ConsumerA), registry.get_key(ConsumerB)],
        )

    def test_get_key(self):
        class ConsumerA(Consumer):
            pass

        class ConsumerB(Consumer):
            pass

        registry = ConsumersRegistry()

        key_a = registry.get_key(ConsumerA)
        self.assertEqual(key_a, f"{ConsumerA.__module__}.{ConsumerA.__name__}")

        key_b = registry.get_key(ConsumerB)
        self.assertEqual(key_b, f"{ConsumerB.__module__}.{ConsumerB.__name__}")



import logging
from pydoc import locate
from typing import Optional

from confluent_kafka import Consumer as ConfluentConsumer
from confluent_kafka import cimpl

from django_kafka.conf import settings
from django_kafka.topic import Topic

logger = logging.getLogger(__name__)


class Topics(dict):
    def __init__(self, *topics: Topic):
        for topic in topics:
            self[topic.name] = topic


class Consumer:
    """
    Available settings of the producers (P) and consumers (C)
        https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
    Consumer configs
        https://kafka.apache.org/documentation/#consumerconfigs
    Kafka Client Configuration
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    confluent_kafka.Consumer API
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-consumer
    """

    topics: Topics[str, Topic]
    config: dict

    polling_freq = settings.POLLING_FREQUENCY
    default_logger = logger
    default_error_handler = settings.ERROR_HANDLER

    def __init__(self, config: Optional[dict] = None, **kwargs):
        kwargs.setdefault("logger", self.default_logger)
        kwargs.setdefault("error_cb", locate(self.default_error_handler)())

        self.config = {
            "client.id": settings.CLIENT_ID,
            **settings.GLOBAL_CONFIG,
            **settings.CONSUMER_CONFIG,
            **getattr(self, "config", {}),
            **(config or {}),
        }

        self._consumer = ConfluentConsumer(self.config, **kwargs)

    def __getattr__(self, name):
        """proxy consumer methods."""
        if name not in {"config"}:
            # For cases when `Consumer.config` is not set and
            #  `getattr(self, "config", {})` is called on `__init__`,
            #  the initialization will fail because `_consumer` is not yet set.
            return getattr(self._consumer, name)
        raise AttributeError(f"'{self.__class__.__name__}' has no attribute 'name'")

    def start(self):
        # define topics
        self.subscribe(topics=list(self.topics))
        while True:
            # poll every self.polling_freq seconds
            if msg := self.poll(timeout=self.polling_freq):
                self.process_message(msg)

    def stop(self):
        self.close()

    def process_message(self, msg: cimpl.Message):
        if msg_error := msg.error():
            self.handle_error(msg_error)
            return

        try:
            self.topics[msg.topic()].consume(msg)
        # ruff: noqa: BLE001 (we do not want consumer to stop if message processing is failing in any circumstances)
        except Exception as error:
            self.handle_error(error)
        else:
            self.commit_offset(msg)

    def commit_offset(self, msg: cimpl.Message):
        if not self.config.get("enable.auto.offset.store"):
            # Store the offset associated with msg to a local cache.
            # Stored offsets are committed to Kafka by a background
            #  thread every 'auto.commit.interval.ms'.
            # Explicitly storing offsets after processing gives at-least once semantics.
            self.store_offsets(msg)

    def handle_error(self, error):
        logger.error(error, exc_info=True)



import logging
from abc import ABC, abstractmethod
from typing import Optional, Type

from confluent_kafka import cimpl
from confluent_kafka.schema_registry.avro import AvroDeserializer, AvroSerializer
from confluent_kafka.serialization import (
    Deserializer,
    MessageField,
    SerializationContext,
    Serializer,
    StringDeserializer,
    StringSerializer,
)

from django_kafka import kafka
from django_kafka.exceptions import DjangoKafkaError

logger = logging.getLogger(__name__)


class Topic(ABC):
    key_serializer: Type[Serializer] = StringSerializer
    value_serializer: Type[Serializer] = StringSerializer

    key_deserializer: Type[Deserializer] = StringDeserializer
    value_deserializer: Type[Deserializer] = StringDeserializer

    @property
    @abstractmethod
    def name(self) -> str:
        """Define Kafka topic name"""

    def consume(self, msg: cimpl.Message):
        """Implement message processing"""
        raise NotImplementedError

    def produce(self, value: any, **kwargs):
        key_serializer_kwargs = kwargs.pop("key_serializer_kwargs", {}) or {}
        value_serializer_kwargs = kwargs.pop("value_serializer_kwargs", {}) or {}
        headers = kwargs.get("headers")

        if "key" in kwargs:
            kwargs["key"] = self.serialize(
                kwargs["key"], MessageField.KEY, headers, **key_serializer_kwargs)

        kafka.producer.produce(
            self.name,
            self.serialize(
                value, MessageField.VALUE, headers, **value_serializer_kwargs,
            ),
            **kwargs,
        )

    def serialize(
        self,
        value,
        field: MessageField,
        headers: Optional[dict | list] = None,
        **kwargs,
    ):
        if field == MessageField.VALUE:
            serializer = self.get_value_serializer(**kwargs)
            return serializer(value, self.context(MessageField.VALUE, headers))

        if field == MessageField.KEY:
            serializer = self.get_key_serializer(**kwargs)
            return serializer(value, self.context(MessageField.KEY, headers))

        raise DjangoKafkaError(f"Unsupported serialization field {field}.")

    def deserialize(
        self,
        value,
        field: MessageField,
        headers: Optional[dict | list] = None,
        **kwargs,
    ):
        if field == MessageField.VALUE:
            deserializer = self.get_value_deserializer(**kwargs)
            return deserializer(value, self.context(MessageField.VALUE, headers))

        if field == MessageField.KEY:
            deserializer = self.get_key_deserializer(**kwargs)
            return deserializer(value, self.context(MessageField.KEY, headers))

        raise DjangoKafkaError(f"Unsupported deserialization field {field}.")

    def get_key_serializer(self, **kwargs):
        return self.key_serializer(**kwargs)

    def get_value_serializer(self, **kwargs):
        return self.value_serializer(**kwargs)

    def get_key_deserializer(self, **kwargs):
        return self.key_deserializer(**kwargs)

    def get_value_deserializer(self, **kwargs):
        return self.value_deserializer(**kwargs)

    def context(
        self,
        field: MessageField,
        headers: Optional[dict | list] = None,
    ) -> SerializationContext:
        return SerializationContext(self.name, field, headers=headers)


class AvroTopic(Topic, ABC):
    """
    Consume.
        Defining schemas is not necessary as it gets retrieved automatically from the Schema Registry.

    Produce.
        Defining `value_schema` is required (`key_schema` is required when using keys).
        It gets submitted to the Schema Registry

    Multiple schemas and one Topic:
        `AvroTopic.produce` takes `serializer_kwargs` kw argument.
        `AvroSerializer` then gets initialized with the provided kwargs.
        When producing you can tell which schema to use for your message:
        ```python
        schema = {
            "type": "record",
            "name": "ValueTest",
            "fields": [
                {"name": "value", "type": "string"},
            ]
        }
        topic.produce({"value": 1}, value_serializer_kwargs={"schema_str": json.dumps(schema)})
        ```

    [Cofluent AvroSerializer Config](https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#avroserializer)
    [Avro schema definition](https://avro.apache.org/docs/1.11.1/specification/)
    """  # noqa: E501
    key_schema: str
    value_schema: str
    schema_config: dict

    def get_key_serializer(self, **kwargs):
        kwargs.setdefault("schema_str", getattr(self, "key_schema", None))
        kwargs.setdefault("conf", getattr(self, "schema_config", None))

        return AvroSerializer(kafka.schema_client, **kwargs)

    def get_value_serializer(self, **kwargs):
        kwargs.setdefault("schema_str", getattr(self, "value_schema", None))
        kwargs.setdefault("conf", getattr(self, "schema_config", None))

        return AvroSerializer(kafka.schema_client, **kwargs)

    def get_key_deserializer(self, **kwargs):
        return AvroDeserializer(kafka.schema_client, **kwargs)

    def get_value_deserializer(self, **kwargs):
        return AvroDeserializer(kafka.schema_client, **kwargs)



import logging
from multiprocessing.pool import Pool
from typing import Optional

from confluent_kafka.schema_registry import SchemaRegistryClient
from django.utils.functional import cached_property
from django.utils.module_loading import autodiscover_modules

from django_kafka.conf import settings
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.producer import Producer
from django_kafka.registry import ConsumersRegistry

logger = logging.getLogger(__name__)

__version__ = "0.1.0"

__all__ = [
    "autodiscover",
    "DjangoKafka",
    "kafka",
]


def autodiscover():
    autodiscover_modules("consumers")


class DjangoKafka:
    consumers = ConsumersRegistry()

    @cached_property
    def producer(self) -> Producer:
        return Producer()

    @cached_property
    def schema_client(self) -> SchemaRegistryClient:
        """
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#schemaregistryclient
        """
        if not settings.SCHEMA_REGISTRY:
            raise DjangoKafkaError(
                "`SCHEMA_REGISTRY` configuration is not defined.",
            )

        return SchemaRegistryClient(settings.SCHEMA_REGISTRY)

    def start_consumer(self, consumer: str):
        self.consumers[consumer]().start()

    def start_consumers(self, consumers: Optional[list[str]] = None):
        consumers = consumers or list(self.consumers)
        with Pool(processes=len(consumers)) as pool:
            try:
                pool.map(self.start_consumer, consumers)
            except KeyboardInterrupt:
                # Stops the worker processes immediately without completing
                #  outstanding work.
                pool.terminate()
                # Wait for the worker processes to exit.
                # Should be called after close() or terminate().
                pool.join()
                logger.debug("KeyboardInterrupt. Pool workers terminated.")
            else:
                # Prevents any more tasks from being submitted to the pool.
                # Once all the tasks have been completed the worker processes will exit.
                pool.close()
                # Wait for the worker processes to exit.
                # Should be called after close() or terminate().
                pool.join()


kafka = DjangoKafka()



from django.apps import AppConfig


class DjangoKafkaConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_kafka"

    def ready(self):
        super().ready()
        self.module.autodiscover()



from django.db import models
from django.utils.translation import gettext_lazy as _


class KafkaSkipQueryset(models.QuerySet):
    def update(self, **kwargs) -> int:
        kwargs.setdefault("kafka_skip", False)
        return super().update(**kwargs)


class KafkaSkipMixin(models.Model):
    """
    For models (tables) which are synced with other database(s) in both directions.

    Every update which happens from within the system should set `kafka_skip=False`,
     global producer (kafka connect, django post_save signal, etc.) will then create
     a new event.

    When db update comes from the consumed event, then the row should be manually
     marked for skip `kafka_skip=True`, and kafka connector or global python producer
     should not generate a new one by filtering it out based on `kafka_skip` field.
    """

    kafka_skip = models.BooleanField(
        _("Kafka skip"),
        help_text=_(
            "Wont generate an event if `True`."
            "\nThis field is used to filter out the events to break the infinite loop"
            " of message generation when synchronizing 2+ databases."
            "\nGets reset to False on .save() method call.",
        ),
        default=False,
    )
    objects = KafkaSkipQueryset.as_manager()

    class Meta:
        abstract = True
        base_manager_name = "objects"

    def save(
        self,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None,
    ):
        if update_fields is None:
            self.kafka_skip = False

        elif "kafka_skip" not in update_fields:
            self.kafka_skip = False
            update_fields = ["kafka_skip", *update_fields]

        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )
